{%- if bindings == "pyo3" -%}
use pyo3::prelude::*;

/// A Python module implemented in Rust.
#[pymodule]
mod {{crate_name}} {
    use pyo3::prelude::*;

    /// Formats the sum of two numbers as string.
    #[pyfunction]
    fn sum_as_string(a: usize, b: usize) -> PyResult<String> {
        Ok((a + b).to_string())
    }
}
{%- elif bindings == "uniffi" -%}
fn add(a: u32, b: u32) -> u32 {
    a + b
}

uniffi::include_scaffolding!("example");
{%- endif %}
