[package]
name = "{{ name }}"
version = "0.1.0"
edition = "2024"

# See more keys and their definitions at https://doc.rust-lang.org/cargo/reference/manifest.html
{%- if bindings != "bin" %}
[lib]
name = "{{ crate_name }}"
crate-type = ["cdylib"]
{%- endif %}

[dependencies]
{% if bindings == "pyo3" -%}
pyo3 = "0.28.3"
{% elif bindings == "uniffi" -%}
uniffi = "0.28.0"

[build-dependencies]
uniffi = { version = "0.28.0", features = ["build"] }
{% endif -%}
